<?php

namespace Teknisa\FIN\Services;

use Zeedhi\Framework\DTO\Request\DataSet;
use Teknisa\Libs\Util\XLS\FileLoader;
use Teknisa\Libs\Util\XLS\FileReader;
use Teknisa\Libs\Util\Environment;
use Teknisa\FIN\Services\ImportacaoTitulosError;
use Teknisa\FIN\Factories\ImportacaoTitulosFactory;
use Teknisa\Libs\Util\XLS\Table;
use Teknisa\Libs\Util\XLS\Exception;
use Zeedhi\Framework\DataSource;
use PHPExcel_Cell;
use PHPExcel_Shared_Date;
use PhpOffice\PhpSpreadsheet;

class ImportacaoTitulosService{ 
    
    const TITULOS_PAGAR = 'P';
    const TITULOS_RECEBER = 'R';
    const IMPORT_EMPTY_CELL_VALUE_ERROR = 'IMPORT_EMPTY_CELL_VALUE_ERROR';
    const IMPORT_NOT_FOUND_CELL_ERROR = 'IMPORT_NOT_FOUND_CELL_ERROR';
    const IMPORT_NOT_RELATED_CELL_ERROR = 'IMPORT_NOT_RELATED_CELL_ERROR';
    const IMPORT_ROW_EXISTS_ERROR = 'IMPORT_ROW_EXISTS_ERROR';
    const IMPORT_FILE_FORMAT_ERROR = 'IMPORT_FILE_FORMAT_ERROR';
    const IMPORT_ROW_CENTCUST_ERROR = 'IMPORT_ROW_CENTCUST_ERROR';
    
    const CDFILIAL_LENGTH = 4;
    const CDREDUCONT_LENGTH = 3;
    const CDTIPOOPER_LENGTH = 2;
    const CDTPCTPAGAR_LENGTH = 6;
    const CDTPCTRECEB_LENGTH = 6;
    
    protected $entityManager;
    protected $environment;
    protected $factory;
    private $titleType;
    protected $columnNames; //nome das colunas da planilha
    protected $optionalColumnNames; //nome das colunas da planilha que aceitam null
    protected $nrorg;
    
    private function existsFieldOnDB($value, $cdempresa, $columnName, $titleType){
        if($titleType === self:: TITULOS_PAGAR){
            return self::existsFieldOnDBPagar($value, $cdempresa, $columnName);
        }
        else{
            if($titleType === self:: TITULOS_RECEBER){
                return self::existsFieldOnDBReceber($value, $cdempresa, $columnName);
            }
        }
        return true;
    }
    
    private function existsFieldOnDBPagar($value, $cdempresa, $columnName){
        switch($columnName){
            case 'FILIAL DE ORIGEM':
                return ($this->factory->getFilialDeOrigem($cdempresa, $value, $this->nrorg));
                break;
            case 'CNPJ / CPF':
                return ($this->factory->getCNPJOrCPFPagar($value, $this->nrorg));
                break;
            case 'TIPO DE OPERAÇÃO':
                return ($this->factory->getTipoDeOperacaoPagar($value, $this->nrorg));
                break;
            case 'TIPO DE CONTA':
                return ($this->factory->getTipoDeContaPagar($value, $this->nrorg));
                break;
            case 'SÉRIE DA NOTA':
                return ($this->factory->getSerieDaNotaPagar($value, $this->nrorg));
                break;
            default:
                return false;
        }
    }
    
    private function existsFieldOnDBReceber($value, $cdempresa, $columnName){
        switch($columnName){
            case 'FILIAL':
                return ($this->factory->getFilialDeOrigem($cdempresa, $value, $this->nrorg));
                break;
            case 'CPF / CNPJ':
                return ($this->factory->getCNPJOrCPFReceber($value, $this->nrorg));
                break;
            case 'TIPO DE OPERAÇÃO':
                return ($this->factory->getTipoDeOperacaoPagar($value, $this->nrorg));
                break;
            case 'TIPO CONTA':
                return ($this->factory->getTipoDeContaReceber($value, $this->nrorg));
                break;
            case 'SÉRIE':
                return ($this->factory->getSerieDaNotaPagar($value, $this->nrorg));
                break;
            case 'COD REDUZIDO CONTA CORRENTE':
                return ($this->factory->getCodReduzidoContaCorrente(NULL, $value, $this->nrorg));
                break;
            case 'TIPO BAIXA':
                return ($this->factory->getTipoDeBaixa($value, $this->nrorg));
                break;
            default:
                return false;
        }
    }
    
    private function fieldNeedAnotherRowData($columnName){
        $columnNames = array('NÚMERO DA NF', 'NÚMERO NOTA');
        return in_array($columnName, $columnNames);
    }
    
    private function formatData($data){
        // Alteração de depreciação do PHPExcel - Nas versões acima de 70400
        if ((PHP_VERSION_ID >= 70400) && (gettype($data) == 'integer'))
            $data =  date('d/m/Y', PhpSpreadsheet\Shared\Date::excelToTimestamp($data + 1));
        elseif(gettype($data) == 'double')
            $data =  date('d/m/Y', PHPExcel_Shared_Date::ExcelToPHP($data + 1.0));
        
        return $data;
    }
    
    private function formatValue($value, $columnName, $titleType, $currentRow = null){
        if($titleType == self::TITULOS_PAGAR){
            switch($columnName){
                case 'CNPJ / CPF':
                    $value = strval($value);
                    break;
                case 'NÚMERO DA NF':
                    $notafiscal = $this->factory->getNumeroNotaFiscalLength($currentRow['FILIAL DE ORIGEM'], $this->nrorg);
                    $value = self::leftZeroFill($value, $notafiscal['NRTAMENTRADA']);
                    break;
                case 'FILIAL DE ORIGEM':
                    $value = self::leftZeroFill($value, self::CDFILIAL_LENGTH);
                    break;
                case 'TIPO DE OPERAÇÃO':
                    $value = self::leftZeroFill($value, self::CDTIPOOPER_LENGTH);
                    break;
                case 'TIPO DE CONTA':
                    $value = self::leftZeroFill($value, self::CDTPCTPAGAR_LENGTH);
                    break;
                case 'TOTAL NF':
                case 'VALOR':
                    $value = str_replace(',', '.', $value);
                    $value = floatval($value);
                    break;
                case 'DATA DE EMISSÃO':
                case 'DATA DE ENTRADA':
                case 'DATA DE VENCIMENTO':
                    $value = self::formatData($value);
                    break;
            }
            
        }
        else{
            switch($columnName){
                case 'CPF / CNPJ': 
                    $value = strval($value);
                    break;
                case 'NÚMERO NOTA':
                    $notafiscal = $this->factory->getNumeroNotaFiscalLength($currentRow['FILIAL'], $this->nrorg);
                    $value = self::leftZeroFill($value, $notafiscal['NRTAMSAIDA']);
                    break;
                case 'FILIAL':
                    $value = self::leftZeroFill($value, self::CDFILIAL_LENGTH);
                    break;
                case 'TIPO DE OPERAÇÃO':
                    $value = self::leftZeroFill($value, self::CDTIPOOPER_LENGTH);
                    break;
                case 'TIPO CONTA':
                    $value = self::leftZeroFill($value, self::CDTPCTRECEB_LENGTH);
                    break;
                case 'COD REDUZIDO CONTA CORRENTE':
                    $value = self::leftZeroFill($value, self::CDREDUCONT_LENGTH);
                    break;
                case 'VALOR ORIGINAL':
                case 'VALOR MULTA':
                    $value = str_replace(',', '.', $value);
                    $value = floatval($value);
                    break;
                case 'DT EMISSÃO':
                case 'DT SAÍDA':
                case 'DT VENC':
                    $value = self::formatData($value);
                    break;
            }
        }
        return $value;
    }
    
    private function getTituloColumnNames($titleType){
        $columnNames = array();
        if($titleType == self::TITULOS_PAGAR){
            $columnNames = array(
                'FILIAL DE ORIGEM', 'CNPJ / CPF', 'TIPO DE OPERAÇÃO', 'TIPO DE CONTA',
                'SÉRIE DA NOTA', 'TOTAL NF', 'NÚMERO DA NF', 'DATA DE EMISSÃO',
                'DATA DE ENTRADA', 'DATA DE VENCIMENTO', 'VALOR'
            );
        }
        if($titleType == self::TITULOS_RECEBER){
            $columnNames = array(
                'NOME FANTASIA', 'FILIAL', 'CPF / CNPJ', 'NÚMERO NOTA', 'SÉRIE',
                'DT EMISSÃO', 'DT SAÍDA', 'VALOR ORIGINAL', 'TIPO CONTA', 'DT VENC',
                'VALOR MULTA', 'MORA', 'COD REDUZIDO CONTA CORRENTE', 'NOSSO NÚMERO',
                'CÓDIGO BARRA', 'LINHA DIGITÁVEL', 'TIPO DE OPERAÇÃO', 'DESCRIÇÃO',
                'TIPO BAIXA'
            );
        }
        return $columnNames;
    }
    
    private function getTituloOptionalColumnNames($titleType){
        $optionalColumnNames = array();
        if($titleType == self::TITULOS_PAGAR){
            $optionalColumnNames = array(); 
        }
        if($titleType == self::TITULOS_RECEBER){
            $optionalColumnNames = array(
                'VALOR MULTA', 'MORA', 'COD REDUZIDO CONTA CORRENTE', 'NOSSO NÚMERO',
                'CÓDIGO BARRA', 'LINHA DIGITÁVEL','TIPO BAIXA'
            );  
        }
        return $optionalColumnNames;
    }
    
    private function getTituloSearchColumnNames($titleType){
        $seachColumnNames = array();
        if($titleType == self::TITULOS_PAGAR){
            $seachColumnNames = array(
                'FILIAL DE ORIGEM', 'CNPJ / CPF', 'TIPO DE OPERAÇÃO', 'TIPO DE CONTA', 
                'SÉRIE DA NOTA'
            );
        }
        if($titleType == self::TITULOS_RECEBER){
            $seachColumnNames = array(
                'FILIAL', 'CPF / CNPJ', 'TIPO DE OPERAÇÃO', 'TIPO CONTA', 'SÉRIE',
                'COD REDUZIDO CONTA CORRENTE', 'TIPO BAIXA'
            );
        }
        return $seachColumnNames;
    }
    
    private function isFieldValueOptional($columnName, $titleType){
        return in_array($columnName, self::getTituloOptionalColumnNames($titleType));
    }
    
    private function isFieldToSearch($columnName, $titleType){
        return in_array($columnName, self::getTituloSearchColumnNames($titleType));
    }
    
    private function validateTituloFieldValue($value, $row, $column, $columnName, $titleType, $dbColumnValues){
        $error = NULL;
        if($value === NULL && !self::isFieldValueOptional($columnName, $titleType)){
            $error = new ImportacaoTitulosError(self::IMPORT_EMPTY_CELL_VALUE_ERROR, $row, $column, $columnName);
        }
        else{
            if(self::isFieldToSearch($columnName, $titleType)){
                if(!self::existsFieldOnDB($value, $dbColumnValues['CDEMPRESA'], $columnName, $titleType)){
                    $error = new ImportacaoTitulosError(self::IMPORT_NOT_FOUND_CELL_ERROR, $row, $column, $columnName);
                }
                else{
                    if($titleType == self::TITULOS_RECEBER && $columnName == 'COD REDUZIDO CONTA CORRENTE'
                        && $this->factory->getCodReduzidoContaCorrente($dbColumnValues['CDEMPRESA'], $value, $this->nrorg) === NULL){
                        $error = new ImportacaoTitulosError(self::IMPORT_NOT_RELATED_CELL_ERROR, $row, $column, $columnName, 'CDEMPRESA');
                    }
                }
            }
        }
        return $error;
    }
    
    private function leftZeroFill($string, $length){
        $string = str_pad(strval($string), $length, '0', STR_PAD_LEFT); 
        return $string;
    }
    
    private function validateCentCust($tableRow, $rowIndex, $dbColumnValues, $titleType){
        $error = null;
        if($titleType == self::TITULOS_PAGAR){
            $cdfilial = $tableRow['FILIAL DE ORIGEM'];
        }
        else{
            $cdfilial = $tableRow['FILIAL'];
        }
        $centcust = $this->factory->buscaCentCustFili($cdfilial, $dbColumnValues['CDEMPRESA'], $this->nrorg);
        if(!$centcust){
            $error = new ImportacaoTitulosError(self::IMPORT_ROW_CENTCUST_ERROR, $rowIndex);
        }
        return $error;
    }
    
    private function validateNotaFiscal($tableRow, $row, $titleType){
        $error = null;
        if($titleType == self::TITULOS_PAGAR){
            $cdfilirece = $tableRow['FILIAL DE ORIGEM'];
            $nrnotafisc = $tableRow['NÚMERO DA NF'];
            $cdserieentr = $tableRow['SÉRIE DA NOTA'];
            $nrinsjurnf = $tableRow['CNPJ / CPF'];
            $vrnotafisc = $tableRow['TOTAL NF'];
            $notafiscal = $this->factory->getNotaFicalPagar($cdfilirece, $nrnotafisc, $cdserieentr, $nrinsjurnf, $vrnotafisc, $this->nrorg);
        }
        else{
            $cdfiliemit = $tableRow['FILIAL'];
            $nrnotafisc = $tableRow['NÚMERO NOTA'];
            $cdserie = $tableRow['SÉRIE'];
            $nrinsjurnf = $tableRow['CPF / CNPJ'];
            $vrnotafisc = $tableRow['VALOR ORIGINAL'];
            $notafiscal = $this->factory->getNotaFicalReceber($cdfiliemit, $nrnotafisc, $cdserie, $nrinsjurnf, $vrnotafisc, $this->nrorg);
        }

        if($notafiscal){
            $error = new ImportacaoTitulosError(self::IMPORT_ROW_EXISTS_ERROR, $row);
        }
        return $error;
    }
    
    public function __construct($entityManager, Environment $environment, ImportacaoTitulosFactory $factory) {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->factory = $factory;
        $this->nrorg = $this->environment->getNrOrgTrab();
    }
    
    public function buildTitulosAPagarDataSet($cdempresa, $tableRows){
        $dataSet = array();
        foreach($tableRows as $tableRow){
            array_push($dataSet, self::buildTitulosAPagarRow($cdempresa, $tableRow['content']));
        }
        return new DataSet(new DataSource\DataSet(null, $dataSet), null, null, null);
    }
    
    public function buildTitulosAPagarRow($cdempresa, $worksheetRow){
        $row = array();
        $row['NRORG'] = $this->nrorg;
        //insere empresa
        $row['CDEMPRESARECE'] = $cdempresa;
        
        //insere dados da filial
        $filial = $this->factory->getFilialDeOrigem($cdempresa, $worksheetRow['FILIAL DE ORIGEM'], $this->nrorg);
        $row['CDFILIAL'] = $this->environment->getCdFilial();
        $row['CDFILIRECE'] = $filial['CDFILIAL'];
        $row['CDFILIALRECE'] = $filial['CDFILIAL'];
        $row['CDINSCESTARECE'] = $filial['CDINSCESTA'];
        $row['CDINSCMUNIRECE'] = $filial['CDINSCMUNI'];
        
        //insere dados do fornecedor
        $fornecedor = $this->factory->getCNPJOrCPFPagar($worksheetRow['CNPJ / CPF'], $this->nrorg);
        $row['CDFORNECED'] = $fornecedor['CDFORNECED'];
        $row['NRINSJURFORN'] = $fornecedor['NRINSJURFORN'];
        $row['IDTPIJURFORN'] = $fornecedor['IDTPIJURFORN'];
        
        //insere dados nota fiscal
        $row['NRNOTAFISC'] = $worksheetRow['NÚMERO DA NF'];
        $row['CDSERIE'] = $worksheetRow['SÉRIE DA NOTA'];
        $row['IDSTATUS'] = 'A';
        $row['CDTIPOOPER'] = $worksheetRow['TIPO DE OPERAÇÃO'];
        $row['VRNOTAFISC'] = $worksheetRow['TOTAL NF'];
        
        //insere dados da nota
        $tipoConta = $this->factory->getTipoDeContaPagar($worksheetRow['TIPO DE CONTA'], $this->nrorg);
        $row['CDTPCTPAGAR'] = $tipoConta['CDTPCTPAGAR'];
        $row['DSCONTAPAG'] = $tipoConta['DSTPCTPAGAR'];
        $row['CDCLASSFINA'] = $tipoConta['CDCLASSFINA'];
        $row['VRCONTA'] = $worksheetRow['VALOR'];
        
        //insere dados operador
        $row['CDOPERACTPG'] = $this->environment->getCdOperador();
        
        //insere datas
        $dataEmissao = $worksheetRow['DATA DE EMISSÃO'];
        $dataEntrada = $worksheetRow['DATA DE ENTRADA'];
        $dataVencimento = $worksheetRow['DATA DE VENCIMENTO'];
        $row['DTATUAVENPAG'] = $dataVencimento;
        $row['DTEMISSAO'] = $dataEmissao;
        $row['DTENTRSAID'] = $dataEntrada;
        
        //Dado para informar o parcelamento
        $row['PARCELAMENTO'] = true;
        
        return $row;
    }
    
    public function buildTitulosAReceberDataSet($cdempresa, $tableRows){
        $dataSet = array();
        foreach($tableRows as $tableRow){
            array_push($dataSet, self::buildTitulosAReceberRow($cdempresa, $tableRow['content']));
        }
        return new DataSet(new DataSource\DataSet(null, $dataSet), null, null, null);
    }
    
    public function buildTitulosAReceberRow($cdempresa, $worksheetRow){
        $row = array();
        $row['NRORG'] = $this->nrorg;
        
        //insere dados da filial
        $filial = $this->factory->getFilialDeOrigem($cdempresa, $worksheetRow['FILIAL'], $this->nrorg);
        $row['CDFILIAL'] = $this->environment->getCdFilial();
        $row['CDFILIEMIT'] = $filial['CDFILIAL'];
        $row['CDEMPRESAEMIT'] = $cdempresa;
        $row['CDINSCESTAEMIT'] = $filial['CDINSCESTA'];;
        $row['CDINSCMUNIEMIT'] = $filial['CDINSCMUNI'];;
        
        //insere dados do cliente
        $cliente = $this->factory->getCNPJOrCPFReceber($worksheetRow['CPF / CNPJ'], $this->nrorg);
        $row['CDCLIENTE'] = $cliente['CDCLIENTE'];
        $row['NRINSJURCLIE'] = $cliente['NRINSJURCLIE'];
        $row['IDTPIJURCLIE'] = $cliente['IDTPIJURCLIE'];
        $row['NRLICENCAUSO'] = null;
        
        //insere dados nota fiscal
        $row['NRNOTAFISC'] = $worksheetRow['NÚMERO NOTA'];
        $row['CDSERIE'] = $worksheetRow['SÉRIE'];
        $row['IDSTATUS'] = 'A';
        $row['CDTIPOOPER'] = $worksheetRow['TIPO DE OPERAÇÃO'];
        $row['CDVENDEDOR'] = null;
        $row['VRNOTAFISC'] = $worksheetRow['VALOR ORIGINAL'];
        
        //insere dados da nota
        $tipoConta = $this->factory->getTipoDeContaReceber($worksheetRow['TIPO CONTA'], $this->nrorg);
        $row['CDTPCTRECEB'] = $tipoConta['CDTPCTRECEB'];
        $row['DSCONTAREC'] = $tipoConta['DSTPCTRECEB'];
        $row['CDCLASSFINA'] = $tipoConta['CDCLASSFINA'];
        
        //insere datas
        $dataEmissao = $worksheetRow['DT EMISSÃO'];
        $dataSaida = $worksheetRow['DT SAÍDA'];
        $dataVencimento = $worksheetRow['DT VENC'];
        $row['DTEMISSAO'] = $dataEmissao;
        $row['DTENTRSAID'] = $dataSaida;
        $row['DTATUAVENREC'] = $dataVencimento;
        
        //insere dados do boleto
        $row['VRMULTABREC'] = $worksheetRow['VALOR MULTA'];
        $row['VRMORDIABREC'] = $worksheetRow['MORA'];
        $row['CDREDUCONT'] = $worksheetRow['COD REDUZIDO CONTA CORRENTE'];
        $row['CDNOSNUMREC'] = $worksheetRow['NOSSO NÚMERO'];
        $row['CDBARRAREC'] = $worksheetRow['CÓDIGO BARRA'];
        $row['DSLINHABARREC'] = $worksheetRow['LINHA DIGITÁVEL'];
        $row['DSCONTAREC'] = $worksheetRow['DESCRIÇÃO'];
        $row['CDTPBAIXAREC'] = $worksheetRow['TIPO BAIXA'];
         
        //Dado para informar o parcelamento
        $row['PARCELAMENTO'] = true;
        
        return $row;
    }
    
    public function formatTitulos(Table $table, $titleType){
        $columnsReference = $table->getColumnsReference();
        $tableRows = $table->getRows();
        $columnsNumber = $table->getNumberOfColumns();
        foreach ($tableRows as &$row) {
            for($column = 0; $column < $columnsNumber; $column++){
                // Alteração de depreciação do PHPExcel - Nas versões acima de 70400
                if(PHP_VERSION_ID >= 70400)
                     $columnLetterIndex = PhpSpreadsheet\Cell\Coordinate::stringFromColumnIndex($column + 1);
                else $columnLetterIndex = PHPExcel_Cell::stringFromColumnIndex($column);
                
                $columnName = $columnsReference[$columnLetterIndex];
                if(!self::fieldNeedAnotherRowData($columnName)){
                    $row['content'][$columnName] = self::formatValue($row['content'][$columnName], $columnName, $titleType); 
                }
                else{
                    $row['content'][$columnName] = self::formatValue($row['content'][$columnName], $columnName, $titleType, $row['content']);
                }
            }
        }
        return $tableRows;
    }
    
    public function readTitulosTable($row, $titleType){
        $file = $row['FILE'];
        $sheetsToLoad = array('Plan1');
        $xlsFile = FileLoader::loadFromBase64($file['content'], $file['name'], $sheetsToLoad);
        if(!$xlsFile){
            throw new \Exception(self::IMPORT_LOAD_FILE_ERROR);
        }
        $worksheet = $xlsFile->getActiveSheet();
        $columnNames = self::getTituloColumnNames($titleType);
        $table = FileReader::readWorksheetTable($worksheet, $columnNames);
        return $table;
    }
    
    public function validateTitulosTable(Table $table, $titleType, $dbColumnValues){
        $errors = array();
        $columnsReference = $table->getColumnsReference();
        $tableRows = $table->getRows();
        $columnsNumber = $table->getNumberOfColumns();
        foreach ($tableRows as $row) {
            for($column = 0; $column < $columnsNumber; $column++){
                $columnLetterIndex = PHPExcel_Cell::stringFromColumnIndex($column);
                $columnName = $columnsReference[$columnLetterIndex];
                $value = $row['content'][$columnName];
                $error = self::validateTituloFieldValue($value, $row['index'], $columnLetterIndex, $columnName, $titleType, $dbColumnValues);
                if($error !== NULL){
                    array_push($errors, $error->toArray());
                }
            }
            $error = self::validateNotaFiscal($row['content'], $row['index'], $titleType);
            if($error !== NULL) array_push($errors, $error->toArray());
            $error = self::validateCentCust($row['content'], $row['index'], $dbColumnValues, $titleType);
            if($error !== NULL) array_push($errors, $error->toArray());
        }
        if(count($errors) > 0){
            throw new Exception($errors, self::IMPORT_FILE_FORMAT_ERROR);
        }
    }
    
}